#! /usr/local/perl
#
#  Convert static IP address /netmask (dot notation) to "define.h"/"ip.inc"
#

sub usage {
    print STDERR "usage: $ego [-v] [-define] [-bat] [-gw] x.y.z.w\n";
    print STDERR "       $ego [-v] [-define] [-bat] [-gw] \@file\n";
    print STDERR "       $ego [-v] [-define] [-bat] [-gw] - < file\n";
    exit 1;
}

$ego = $0;
$ego =~ s/.*[\/\\]//;

$define = 0;
$bat = 0;
$net = 0;
$isgw = 0;
$verbose = 0;
while ($ARGV[0] =~ /^-/ && $ARGV[0] ne '-') {
    $opt = shift;
    if ($opt eq '-define') {
        $define = 1;
    } elsif ($opt eq '-bat') {
        $bat = 1;
    } elsif ($opt eq '-v') {
        $verbose = 1;
    } elsif ($opt eq '-gw') {
        $isgw = 1;
    } else {
        &usage;
    }
}

$ip = shift;
&usage if ($ip eq '');

($ip, $netmask) = &getip($ip);
exit(1) if ($ip eq '');
###print STDERR "ip=$ip, nm=$netmask\n";

(@a) = split(/\./,$ip) ;
$ip = sprintf("%d.%d.%d.%d", $a[0], $a[1], $a[2], $a[3]);

(@b) = split(/\./,$netmask) ;
$netmask = sprintf("%d.%d.%d.%d", $b[0], $b[1], $b[2], $b[3]);

if ($define) {
    printf("#define STATIC_IP_MSB 0x%02x%02x\t/* static IP $ip */\n", $a[0], $a[1]);
    printf("#define STATIC_IP_LSB 0x%02x%02x\n", $a[2], $a[3]);
    printf("#define STATIC_NM_MSB 0x%02x%02x\t/* static IP $netmask */\n", $b[0], $b[1]);
    printf("#define STATIC_NM_LSB 0x%02x%02x\n", $b[2], $b[3]);
} elsif ($bat) {
    print "set PIC=$ip\n";
    print "set NM=$netmask\n";
} else {
    printf("\t.word\t0x%02x%02x\t; IP $ip\n", $a[1], $a[0]);
    printf("\t.word\t0x%02x%02x\n", $a[3], $a[2]);
    printf("\t.word\t0x%02x%02x\t; netmask $netmask\n", $b[1], $b[0]);
    printf("\t.word\t0x%02x%02x\n", $b[3], $b[2]);
}
if ($verbose) {
    if ($netmask eq '0.0.0.0') {
        print STDERR "$ip\n";
    } else {
        print STDERR "$ip/$netmask\n";
    }
}

#
#  Resolve IP address
#
sub getip {
    local($f) = @_;
    local($s, $ip, $nm);

    $ip = '';
    $nm = '0.0.0.0';
    if ($isgw) {
        $ip = $ENV{'PW_GW'} if ($ENV{'PW_GW'} ne '');
    } else {
        $ip = $ENV{'PW_IP'} if ($ENV{'PW_IP'} ne '');
    }

    if ($ip ne '') {
        ;
    } elsif ($f =~ /^[@]/ || $f eq '-') {
        if ($f ne '-') {
            $f =~ s/^[@]//;
            if (!open(IP, "<$f")) {
                print STDERR "Can't open IP address file \"$f\"!\n";
                return '';
            }
        }
        while(1) {
            if ($f ne '-') {
                last if (!($s = <IP>));
            } else {
                last if (!($s = <>));
            }
            ##print STDERR "$s";
            $s =~ s/[\s\r\n]+$//;
            $s =~ s/\s*[#].*$//;
            next if ($s eq '');
            if ($s =~ /^\s*(\d+[.]\d+[.]\d+[.]\d+)[.\r\n\s\d]*$/) {
                $ip = $s;
            }
        }
    } else {
        $ip = $f;
    }
    ##print STDERR "IP/NM=$ip\n";
    if ($ip =~ /^(\d+[.]\d+[.]\d+[.]\d+)\s+(\d+[.]\d+[.]\d+[.]\d+)$/) {
        return ($1, $2);
    } elsif ($ip =~ /^(\d+[.]\d+[.]\d+[.]\d+)$/) {
        return ($1, '0.0.0.0');
    } else {
        print STDERR "IP address and/or netmask ($ip) not of the form x.y.z.w!\n";
        return ('', '');
    }
}
