#! /usr/local/perl
#
#   Load PicoWeb EEPROM using network
#
###print STDERR "$0 @ARGV\n";
sub usage {
    print STDERR "usage: $ego IPaddr [-linux] EEPROM_file [blk_size]\n" ;
    print STDERR "       $ego IPaddr [-linux] -url URL\n" ;
    print STDERR "       $ego IPaddr [-linux] -port port loc byte1 ... byteN\n" ;
    print STDERR "\@IPaddr indicates name of file with IP address" ;
    print STDERR "-port loads SEEPROM bytes (loc/byteN in hex)\n" ;
    exit 1 ;
}

$debug = 0;
$ego = $0 ;
$ego =~ s/^.*[\/\\]// ;
&usage if ($ARGV[0] eq '') ;
$tmp = 'img.tmp';
$tmp2 = 'img2.tmp';
$tmp3 = 'img3.tmp';
$tmp4 = 'img4.tmp';
$is_win = 1; $is_win = 0 if ($ENV{'HOSTTYPE'} =~ /linux/i) ;

$connect_timeout = 6 ;
$read_timeout = 8 ;

$dir = $0 ;
if ($dir =~ /[\/\\]/) {
    $dir =~ s/([\\\/])[^\\\/]+$/$1/;  # use httpget in our directory
} else {
    $dir = './'
}

$status = 0 ;
$url = '' ;
$port = 0 ;
select((select(STDOUT), $| = 1)[0]);    # unbuffer output
select((select(STDERR), $| = 1)[0]);    # unbuffer output

($ip, $netmask) = &getip($ARGV[0]);
exit(1) if ($ip eq '');
shift ;

while ($ARGV[0] =~ /^-/) {
    if ($ARGV[0] eq '-d') {
        $debug = 1;
        shift ;
    }
    if ($ARGV[0] eq '-linux') {
        $is_win = 0;
        shift ;
    }
    if ($ARGV[0] eq '-url') {
        shift ;
        $url = $ARGV[0] ;
        shift ;
    }
    if ($ARGV[0] eq '-port') {
        shift ;
        $port = $ARGV[0] ;
        shift ;
    }
}

if ($is_win) {
    $dir =~ s/\//\\/g ;
    $httpget = $dir."httpget -C $connect_timeout -t $read_timeout";
} else {
    $httpget = $dir."httpget -C $connect_timeout -t $read_timeout";
}
print STDERR "httpget=$httpget\n" if ($debug);

if ($port > 0) {
    &do_open ;
    $addr = hex($ARGV[0]) ;
    printf(STDERR "loc=%04x:", $addr) if ($debug);
    syswrite(I,pack("S",$addr),2) ;
    for ($i=1 ; $i<=$#ARGV ; $i++)  {
        syswrite(I,pack("C",hex($ARGV[$i])),1) ;
        print STDERR " $ARGV[$i]" if ($debug);
    }
    close(I) ;
    print STDERR "\n" if ($debug) ;

    unlink($tmp2) ;
    $cmd = "$httpget -rP $tmp -o $tmp2 $ip:$port" ;
    print STDERR "$cmd\n" if ($debug);
    $out = `$cmd` ;
    print $out if ($debug) ;
    &check_output;
    &cleanup_and_exit;
}

if ($url) {
    &usage if ($ARGV[0] ne '') ;
    $cmd = "$httpget $ip$url" ;
    print STDERR "$cmd\n" if ($debug);
    system($cmd) ;
    exit(0) ;
}

&usage if ($ARGV[1] ne '') ;
&do_open ;
&do_open_tmp($tmp3);
open(F,"<$ARGV[0]") || die "Can't open $ARGV[0]" ;
binmode(F) ;
$addr = 0 ;
if ($ARGV[0] =~ /\.el/i)  {
    sysread(F,$buf,2) ;
    $addr = unpack("S",$buf) ;  # starting address
}
$blocksize = $ARGV[1] ;
$blocksize = 512 if ($blocksize == 0) ; # 256-4-6-6-2-20-20-2(length)
#
# blocksize can still be specified, but it better be a multiple of
# the page size in serial eeprom (64 bytes) - otherwise page wrap will
# occur, and the results will not be favorable!
#
$raw_data = '';
while (sysread(F,$buf,$blocksize) > 0)  {
     ###if (length($buf) != $blocksize)  {  # truncate
        ###close(I) ;
        ###&do_open ;
    ###}
    ###if ($is_win) {
        ###seek(I,0,0) ;
    ###} else {
        ###eval 'sysseek(I,0,0);' ;
        ###seek(I,0,0) if ($@ ne '');
    ###}
    syswrite(I,pack("S",$addr),2) ;
    syswrite(II,pack("S",$addr),2) ;
    ####print "$addr ".length($buf)."\n";
    syswrite(I,$buf,length($buf)) ;
    $raw_data .= $buf;
    syswrite(II,pack("S",length($buf)),2) ;
    $addr += length($buf) ;
}
close(I) ;
close(II) ;

# check send commands to write SEEPROM
$len = $blocksize + 2 ;
unlink($tmp2) ;
$cmd = "$httpget -rP $tmp -N $len -o $tmp2 $ip:611" ;
printf(STDERR "%04x: $cmd -> ", $addr) if ($debug);
##printf "%04X ",$addr ;
print "load";
##system($cmd) ;
$s = `$cmd` ;
print STDERR "$s\n" if ($debug);
&check_output ;

# check send commands to check SEEPROM
die if (-z $tmp3);
$cmd = "$httpget -rP $tmp3 -N 4 -o $tmp4 $ip:613" ;
printf(STDERR "%04x: $cmd -> ", $addr) if ($debug);
##printf "%04X ",$addr ;
print "read";
##system($cmd) ;
$s = `$cmd` ;
print STDERR "$s\n" if ($debug);
&check_readback ;

&cleanup_and_exit;

sub cleanup_and_exit {
    unlink($tmp) if ($debug == 0) ;
    unlink($tmp2) if ($debug == 0) ;
    unlink($tmp3) if ($debug == 0) ;
    unlink($tmp4) if ($debug == 0) ;
    exit $status ;
}

sub check_readback {
    if ((-z $tmp4) || (!open(I,"<$tmp4"))) {
         print STDERR "Failed to readback SEEPROM from PicoWeb at $ip!\n" ;
         $status = 1 ;
    } else {
        binmode(I) ;
        sysread(I,$buf,64 * 1024);
        close(I) ;
    }
    if ($raw_data ne $buf) {
         print STDERR "SEEPROM verify failed for PicoWeb at $ip!\n" ;
         $status = 1 ;
    }
    unlink($tmp2) if ($debug == 0) ;
}

sub check_output {
    if ((-z $tmp2) || (!open(I,"<$tmp2"))) {
         print STDERR "Failed to connect to PicoWeb at $ip!\n" ;
         $status = 1 ;
    } else {
        while (<I>) {
            $s = $_ ;
            $s =~ s/[\r\n]+$//;
            if (!($s =~ /^\s+[0-9a-fA-F]+$/)) {
                print "$s" ;
                $status = 1 ;
            }
        }
        close(I) ;
    }
    unlink($tmp2) if ($debug == 0) ;
}

sub do_open  {
    if (!open(I,">$tmp")) {
         print STDERR "Can't create $tmp!\n" ;
         exit 1;
    }
    binmode(I) ;
    select(I) ;
    $| = 1 ;
    select(STDOUT) ;
    $| = 1 ;
}

sub do_open_tmp  {
    local($file) = @_;
    if (!open(II,">$file")) {
         print STDERR "Can't create $file!\n" ;
         exit 1;
    }
    binmode(II) ;
    select(II) ;
    $| = 1 ;
    select(STDOUT) ;
    $| = 1 ;
}

#
#  Resolve IP address
#
sub getip {
    local($f) = @_;
    local($s);

    $ip = '';
    $nm = '0.0.0.0';
    if ($isgw) {
        $ip = $ENV{'PW_GW'} if ($ENV{'PW_GW'} ne '');
    } else {
        $ip = $ENV{'PW_IP'} if ($ENV{'PW_IP'} ne '');
    }

    if ($ip ne '') {
        ;
    } elsif ($f =~ /^[@]/ || $f eq '-') {
        if ($f ne '-') {
            $f =~ s/^[@]//;
            if (!open(IP, "<$f")) {
                print STDERR "Can't open IP address file \"$f\"!\n";
                return '';
            }
        }
        while(1) {
            if ($f ne '-') {
                last if (!($s = <IP>));
            } else {
                last if (!($s = <>));
            }
            ##print STDERR "$s";
            $s =~ s/[\s\r\n]+$//;
            $s =~ s/\s*[#].*$//;
            next if ($s eq '');
            if ($s =~ /^\s*(\d+[.]\d+[.]\d+[.]\d+)[.\r\n\s\d]*$/) {
                $ip = $s;
            }
        }
    } else {
        $ip = $f;
    }
    ##print STDERR "IP/NM=$ip\n";
    if ($ip =~ /^(\d+[.]\d+[.]\d+[.]\d+)\s+(\d+[.]\d+[.]\d+[.]\d+)$/) {
        return ($1, $2);
    } elsif ($ip =~ /^(\d+[.]\d+[.]\d+[.]\d+)$/) {
        return ($1, '0.0.0.0');
    } else {
        print STDERR "IP address and/or netmask ($ip) not of the form x.y.z.w!\n";
        return ('', '');
    }
}
