# pwproj2.pl - process PicoWeb project file
#
# this takes an input and an output file and processes all CGI and images in the
# input file into an EEPROM image in the specified output file.
#
$debug = 0 ;
$MAX_BYTES = 8000 ;   # Max. payload of Ethernet GET response packet
$ver_file = '';
$version = '??';
$HDR_LINES = 3;
$CGI_ORG = 512;

$ego = $0;  $ego =~ s/.*[\\\/]//;
sub usage {
    local($why) = @_;

    print STDERR "usage: $ego [options] project_file project.h prefix\n";
    print STDERR "options:\n";
    print STDERR "  -d         - debug mode\n";
    print STDERR "  -v file    - get version info from 'file'\n";
    print STDERR "'project_file' is PicoWeb project definition file\n";
    print STDERR "'project.h' is output file for user defines\n";
    print STDERR "'prefix' is prefix for user-specific .asm files\n";
    print STDERR "$why\n" if ($why ne '');
    exit 1;
}

while ($ARGV[0] =~ /^-/) {
        $opt = $ARGV[0]; shift;
        if ($opt eq '-d') {
        	$debug = 1;
        } elsif ($opt eq '-v') {
        	$ver_file = $ARGV[0]; shift;
        } elsif ($opt eq '-h') {
        	&usage('');
        } else {
        	print STDERR "$ego: unknown option '$opt'!\n";
        	exit 1;
        }
}

$proj_file = $ARGV[0]; shift; &usage('missing project file') if ($proj_file eq '');
$def_file = $ARGV[0]; shift; &usage('missing project.h file') if ($def_file eq '');
$prefix = $ARGV[0]; shift; &usage('missing prefix') if ($prefix eq '');

$pcode_defs_file = "$prefix.def";

$link_script_file = "$prefix.ld";

%avr_def2def = (
    '#avr_asm',   '.text',
    '#avr_reset', '.section reset',
    '#avr_slow',  '.section slow',
    '#avr_fast',  '.section fast'
);

$hdr = "\n\n\n\n\n/* DO NOT EDIT THIS FILE!!!!! (Auto generated by $ego!) */\n\n\n\n\n";

open(PROJ,$proj_file) || die "$ego: can't find project file '$proj_file'\n" ;

open(DEFS,">$def_file") || die "$ego: can't create output defines file '$def_file'\n" ;
print DEFS $hdr;

open(PDEFS,">$pcode_defs_file") || die "$ego: can't create pcode define file '$pcode_defs_file'\n" ;
print PDEFS $hdr;

open(LDS,">$link_script_file") || die "$ego: can't create linker script file '$link_script_file'\n" ;
print LDS $hdr;
print LDS "#include \"$def_file\"\n";
print LDS "#include \"defdefs.h\"\n";

$imgdata = "" ;
$numimages = 0 ;
$imgloc = 0 ;
$curr_avr_sect = '';

if ($ver_file ne '') {
    if (open(V,"<$ver_file"))  {
        $old_version = $version;
        while (<V>) {
            $line = $_;
            if ($line =~ /^[#]define\s+VERSION\s+"([^"]+)"/) {
                $version =$1 ;
            }
        }
        close(V) ;
        if ($old_version eq $version) {
            print "warning: can't find version string in file '$ver_file'!\n";
        }
    } else {
        print "warning: can't open version file '$ver_file'!\n";
    }
}
print "Version: $version\n";

$image_number = 0 ;
$index = 0 ;
$num_asms = 0;
$line_num = 0;
print DEFS "#line 1 \"$proj_file\"\n";
while (<PROJ>)  {
    ++$line_num;
    $line = $_;
    $line =~ s/[\s\r\n]+$//;
    $org_line = $line;
    $line =~ s/\/\/.*//;	# remove // comments
    $line =~ s/\s+$//;
    if ($line =~ /^([#]avr_[a-z]+)$/) {
        if ($avr_def2def{$1} ne '') {
            $curr_avr_sect = $1;
            last;
        } else {
            print "$ego: Bad directive ($1)!\n";
            exit 1;
        }
    } elsif ($line =~ /^[#]/) {
        print DEFS "#line $line_num \"$proj_file\"\n";
        print DEFS "$org_line\n";
    } elsif ($line =~ /^link\s+(\S+)\s+(\S+)$/i) {
        $directive = $1;
        $parm = $2;
        if ($directive =~ /^search$/i) {
            $link_group .= $parm . ' ';
        } elsif ($directive =~ /^add$/i) {
            $link_input .= $parm . ' ';
        } else {
            print "$proj_file($line_num): Bad link directive! ($org_line)\n";
            exit 1;
        }
    } elsif ($line =~ /^([a-z0-9_]+)[.]([a-z0-9]+)$/i) {
        $base = $1;
        $ext = $2;
        ($file,$x) = split(/\s/, $line) ;
        $images{$image_number} = $file ;
        if ($ext =~ /CGI/i) {
            $cgi_map[$index] = $base;
            $index++ ;
        } elsif ($ext =~ /ASM/i) {
            $asms[$num_asms++] = $line;
        } else {
            $base = $file; $base =~ s/.*[\\\/]//;
            $file_def_line{$file} = $line_num;
            $image_number++ ;
        } 
        print DEFS "// $org_line\n";
    } elsif ($line =~ /^([a-z0-9_]+)[.]([a-z0-9]+)\s+(\S+)$/i) {
        $base = $1;
        $ext = $2;
        $content_type = $3;
        ($file,$x) = split(/\s/, $line) ;
        $images{$image_number} = $file ;
        $base = $file; $base =~ s/.*[\\\/]//;
        $file_def_line{$file} = $line_num;
        $content_types{$file} = $content_type;
        $image_number++ ;
        print DEFS "// $org_line\n";
    } else {
        if ($org_line =~ /^(\s*)\/\/(.*)\s*$/) {
            $s = $1;
            $ss = $2;
            $ss =~ s/\/\*/|*/g;
            $ss =~ s/\*\//*|/g;
            $s = "$s/* $ss */\n";
        } elsif ($org_line =~ /^\s*\/\*.*\*\/\s*$/) {
            $s = "$org_line\n";
        } elsif ($org_line =~ /^\s*$/) {
            $s = "\n";
        } else {
            $s = $org_line;
            $s =~ s/^\s+//;
            $s =~ s/\s+$//;
            print "$proj_file($line_num): Bad project file directive! ($s)\n";
            exit 1;
        }
        print DEFS $s;
    }
}
$number_of_images = $image_number ;

&check_avr_def('#avr_asm');
print ASM $hdr;
print ASM "#line $line_num \"$proj_file\"\n";
print ASM "#include \"$def_file\"\n";

print ASM "#line $line_num \"$proj_file\"\n";
print ASM "#include \"version.h\"\n";

print ASM "#line $line_num \"$proj_file\"\n";
print ASM "#include \"main.asm\"\n";

print ASM "#line $line_num \"$proj_file\"\n";
print ASM "#include \"getip.asm\"\n";

##print ASM "#line $line_num \"$proj_file\"\n";
##print ASM "#include \"ether.inc\"\n";

print ASM "#ifdef DEBUGGER\n";
print ASM "#line $line_num \"$proj_file\"\n";
print ASM "#include \"$prefix.cmd\"\n";
print ASM "#endif\n";

###print ASM "#line $line_num \"$proj_file\"\n";
###print ASM "#include \"bootp.asm\"\n";

######print ASM "#line $line_num \"$proj_file\"\n";
######print ASM "#include \"serial.asm\"\n";

print ASM "#line $line_num \"$proj_file\"\n";
print ASM ".section seeprom_kludge\n";
print ASM ".org $CGI_ORG   /* kludge to fix low SEEPROM pcode execution problem!!!!!!!!! */\n\n";

## make CGI strings default to CGI string section
print ASM "#line $line_num \"$proj_file\"\n";
print ASM "string_section seeprom_cgi,seeprom_cgi_strings\n";
print ASM "#line $line_num \"$proj_file\"\n";
print ASM "string_section seeprom_cgi_strings,seeprom_cgi_strings\n";
print ASM "#line $line_num \"$proj_file\"\n";
print ASM "string_section seeprom_cgi_table,seeprom_cgi_strings\n";

print ASM "#line $line_num \"$proj_file\"\n";
###print ASM "#line 1 \"$proj_file"."_HTML_file_table\"\n";
#######print ASM "  pret\n";      #!!!!!!!!!!!!!!!!!
&drop_file_table;

print ASM "#line $line_num \"$proj_file\"\n";
###print ASM "#line 1 \"$proj_file"."_HTML_images\"\n";
&process_images;
print ASM "#line $line_num \"$proj_file\"\n";

print ASM "#line $line_num \"$proj_file\"\n";
if ($num_asms) {
    for ($i = 0; $i < $num_asms; ++$i) {
        print ASM "#include \"$asms[$i]\"\n";
    }
    print ASM "\n";
}
print ASM "#line $line_num \"$proj_file\"\n";

&check_avr_def($curr_avr_sect);
&new_avr_section;

sub new_avr_section {
    print ASM "$avr_def2def{$curr_avr_sect};   ; begin section $curr_avr_sect\n";
    $section_label = $curr_avr_sect;
    $section_label = substr($section_label, 1);
    print ASM "#line $line_num \"$proj_file\"\n";
    print ASM $section_label."::\n";
}

while (<PROJ>)  {
    ++$line_num;
    $line = $_;
    $line =~ s/[\s\r\n]+$//;
    if ($line =~ /^([#]avr_[a-z]+)$/) {
        if ($avr_def2def{$1} ne '') {
            $curr_avr_sect = $1;
            if ($last_avr_directive ne $curr_avr_sect) {
                &check_avr_def($curr_avr_sect);
            }
            &new_avr_section;
            next;
        } else {
            print "$ego: Bad directive ($1)!\n";
            exit 1;
        }
    } elsif ($line =~ /^\.eseg(.*)$/i) {
        $line = '.section seeprom_cgi' . $1;
    } elsif ($line =~ /^\.cseg(.*)$/i) {
        $line = $avr_def2def{$curr_avr_sect} . $1;
    } elsif ($line =~ /^\.text(.*)$/i) {
        $line = $avr_def2def{$curr_avr_sect} . $1;
    } elsif ($line =~ /^\s*([a-z0-9_]+)[:]\s+pcode_routine\s+(\d+)/i) {
        print PDEFS "$1\tpcode_routine\t$2\n";
    }
    print ASM "$line\n";
}
close(PROJ) ;
close(PDEFS) ;

print LDS "INPUT ( $link_input )\n" if ($link_input ne '');
print LDS "GROUP ( $link_group )\n" if ($link_group ne '');
print LDS "#include \"picoweb.ld\"\n";
close(LDS) ;

print ASM "avr_asm_end"."::\n";
print ASM "#line $line_num \"$proj_file\"\n";
###print ASM "#line 1 \"$proj_file"."_HTML_files\"\n";
&drop_files;

print ASM "#line $line_num \"$proj_file\"\n";
###print ASM "#line 1 \"$proj_file"."_HTML_strings\"\n";
&drop_strings;

##close(OUT) ;
close(DEFS) ;
close(ASM) ;
exit 0;

sub process_images {
    for ($image_number=0 ; $image_number < $number_of_images ; $image_number++)  {
        $key = $file = $images{$image_number} ;
        $key =~ s/[.]/_/g;
        $flag = 0 ;
        $content_type = $content_types{$file} ;
        if ($file =~ /\.jpg$/i)  {
            $flag = 0 ;
            $content_type = 'image/jpeg' if ($content_type eq '');
        }
        elsif ($file =~ /\.gif$/i)  {
            $flag = 0 ;
            $content_type = 'image/gif' if ($content_type eq '');
        }
        elsif ($file =~ /\.png$/i)  {
            $flag = 0 ;
            $content_type = 'image/png' if ($content_type eq '');
        }
        elsif ($file =~ /\.htm$/i)  {
            $flag = 0x4000 ;
            $content_type = 'text/html' if ($content_type eq '');
        }
        elsif ($file =~ /\.html$/i)  {
            $flag = 0x4000 ;
            $content_type = 'text/html' if ($content_type eq '');
        }
        elsif ($file =~ /\.js$/i)  {
            $flag = 0x4000 ;
            $content_type = 'text/html' if ($content_type eq '');
        }
        elsif ($file =~ /\.txt$/i)  {
            $flag = 0x4000 ;
            $content_type = 'text/plain' if ($content_type eq '');
        }
        elsif ($file =~ /\.class$/i)  {
            $flag = 0 ;
            $content_type = 'application/octet-stream' if ($content_type eq '');
        }
        elsif ($file =~ /\.cla$/i)  {
            $flag = 0 ;
            $content_type = 'application/octet-stream' if ($content_type eq '');
        }
        elsif ($content_type eq '') {
            print "$ego: unknown file type ($file) in PicoWeb project!\n" ;
            exit 1 ;              # not an image we know, blow it off
        }
        $ver_pat = '$$VERSION$$';
        $name = $file ;
        open(F,"<$file") || die "can't open project file \"$file\"\n" ;
        binmode(F) ;
        sysread(F,$str,16384) ; # haul in the whole thing
        $len = length($str);
        ###printf STDOUT "ii%02X: $name ($len bytes)\n",$numimages ;
        if ($len > $MAX_BYTES) {
            print "Warning! Truncating '$file' from $len to $MAX_BYTES bytes!\n" ;
            $str = substr($str, 0, $MAX_BYTES) ;
        }
        if ($flag == 0x4000)  {
            # HTML code...preprocess

            # process `t (by removing it!)
            $str =~ s/^\s*[`]t([^a-zA-Z0-9_])/$1/g;

            if ($content_type !~ /^none$/i) {
                $str = "$content_type\n\n" . $str;
                $str = "HTTP/1.1 200 OK\nContent-Type: " . $str;
            }

            $str =~ s/[\r]//g ;
            $str =~ s/[\n]/\r/g ;
            while ($str =~ s/^\r\/\/[^\r]*\r/\r\r/g) { ; } # simple // comments (column 1)
            while ($str =~ s/\r\/\/[^\r]*\r/\r\r/g) { ; } # simple // comments (column 1)
            ##$str =~ s/([^:])\/\/[^\r]*\r/$1\r/g ; # simple // comments

            ###$str =~ s/[$][$]VERSION[$][$]/$version/g ;
            while(1) {
                $i = index($str, $ver_pat);
                if ($i >= 0) {
                    $str = substr($str,0,$i).$version.substr($str,$i+length($ver_pat));
    	    } else {
                    last;
                }
            }

            # put back new-lines
            $str =~ s/[\r]/\n/g ;

    	    print "\n-------  $file  ------\n$str" if ($debug) ;
            ####$str .= pack("C",0) ;   # make sure HTML has a trailing NUL
            &new_string(1, $file, 1-$HDR_LINES, $str);
        }
        elsif ($flag == 0) {
            # no processing...images
            if ($content_type !~ /^none$/i) {
                $str = "$content_type\n\n" . $str;
                $str = "HTTP/1.1 200 OK\nContent-Type: " . $str;
            }
            &new_string(2, $file, 1-$HDR_LINES, $str);
        }
        $imgdata .= $str ;      # append the data
        $offsets{$numimages} = $imgloc ;
        $lengths{$numimages} = length($str) ;
        $flags{$numimages} = $flag ;
        $numimages++ ;
        $imgloc += length($str) ;
        close(F) ;
    }

    print "$numimages images\n" if ($debug);
    print "$imgloc bytes of HTML code/images total\n";
}

sub check_avr_def {
    local($avr_directive) = @_ ;
    local($s, $def) ;

    return if ($avr_directive eq $last_avr_directive);
    $def = $avr_def2def{$avr_directive};
    $asm_file = "$prefix.asm";
    if ($asm_file eq '') {
        die "$ego: internal error!  can't find file for $avr_directive";
    }
    if ($last_asm_file ne $asm_file) {
        if ($last_avr_directive ne '') {
            $s = '>';
            close(ASM)
        }
        open(ASM,"$s>$asm_file") || die "$ego: Can't create output ASM file '$asm_file'\n" ;
        $last_asm_file = $asm_file;
    }
    if ($last_avr_directive ne '') {
        print ASM "#line $line_num \"$proj_file\"\n";
    }
    if (!$avr_defs{$def}) {
        ++$avr_defs{$def};
    }
    $last_avr_directive = $avr_directive;
}

sub new_string {
    local($print, $file, $line_num, $s) = @_;
    local($i, $ss, $label, $key, $len);

    if ($print != 2) {
        while (1) {
            ($ss, $line_num) = &process_html_tags($file, $line_num, $s);
            last if ($ss eq $s);
            $s = $ss;
        }
    }


    $key = $file;
    $key =~ s/[.]/_/g;
    $numstrings += 0;
    $label = "___$key"."_$numstrings";
    $len = length($s);
    $spool_lables[$numstrings] = $label;
    $spool[$numstrings] = $s;
    $slengths{$label} = $len;
    ++$numstrings;

    $i = $line_num; $i = 1 if ($line_num < 0);
    if ($print) {
        $pcode{$file} .= "#line $i \"$file\"\n";
        $pcode{$file} .= "    pprintse $label,$len\n";
        return $label;
    } else {
        $ss = "#line $i \"$file\"\n";
        $ss .= "    pprintse $label,$len\n";
        return $ss;
    }
}

sub drop_files {
    local($key, $len);

    print ASM "; file system pcode\n\n";
    print ASM ".section seeprom_cgi_strings\n";
    for ($i=0 ; $i<$numimages ; $i++)  {
        $key = $file = $images{$i};
        $key =~ s/[.]/_/g;

        &print_file_def_line($file);
        print ASM ";\n; pcode for file: $file\n;\n";
        print ASM "\t.align 1\n";
        print ASM "\t.global _pcode_$key\n";
        print ASM "_pcode_$key:\n";
        print ASM $pcode{$file};
        print ASM "    pret\n\n";
    }
    print ASM $pcode_common;
}

sub drop_strings {
    print ASM "; file system strings\n\n";
    print ASM ".section seeprom_cgi_strings\n";
    for ($i=0 ; $i<$numstrings ; $i++)  {
        $label = $spool_lables[$i];
        $s = $spool[$i];
        &dump_string($label, $s);
    }
}

sub dump_string {
    local($key, $s) = @_;
    local($len, $ss, $ix, $ch, $hex, $hex2);

    $len = length($s);
    print ASM "\t.global $key\n";
    print ASM "$key:\t; $len bytes\n";
    $ss = '';
    for ($ix = 0; $ix < $len; $ix += 2) {
        if (($ix % 10) == 0) {
            print ASM " ; |$ss|\n" if ($ss ne '');
            print ASM " .word ";
            $ss = '';
        } else {
            print ASM ",";
        }
        $ch = substr($s, $ix, 1);
        $hex = ord($ch);
        $ch = '|' if ($ch eq '\n');
        $ch = ' ' if ($ch eq '\'');
        $ch = ' ' if ($ch eq '"');
        $ch = ' ' if ($ch eq '*');
        $ch = ' ' if ($hex < 0x20 || $hex >= 0x7f);
        $ss .= $ch;
        $ch = substr($s, $ix+1, 1);
        if ($ch eq '') {
            $ch = ' ';
            $hex2 = 0;
        } else {
            $hex2 = ord($ch);
        }
        $ch = '|' if ($ch eq '\n');
        $ch = ' ' if ($ch eq '\'');
        $ch = ' ' if ($ch eq '"');
        $ch = ' ' if ($ch eq '*');
        $ch = ' ' if ($hex2 < 0x20 || $hex2 >= 0x7f);
        $ss .= $ch;
        $hex = ($hex2 << 8) | $hex;
        printf(ASM "0x%04x", $hex & 0xffff);
    }
    print ASM " ; |$ss|\n" if ($ss ne '');
    print ASM "; end of _file_$key\n\n";
    print ASM "\n";
}

sub drop_file_table {
    print ASM "; file system root\n";
    print ASM "\n.section seeprom_cgi_table\n";
    print ASM "    .align 1\n";
    print ASM "    .global pfindfile\n";
    print ASM "#line $line_num \"$proj_file\"\n";
    print ASM "pfindfile:\n";
    ###print ASM "    pserial_debug\n";
    ###print ASM "    pprint \"hello world!\\r\\n\"\n";
    for ($image_number=0 ; $image_number < $number_of_images ; $image_number++)  {
        $key = $file = $images{$image_number};
        $key =~ s/[.]/_/g;
        if ($image_number == 0) {
            &print_file_def_line($file);
            print ASM "    pchkfile \"/\",_pcode_$key\n";
        }
        ###print ASM "    pprint \"checking /$file\\r\\n\"\n";
        &print_file_def_line($file);
        print ASM "    pchkfile \"/$file\",_pcode_$key\n";
    }
    print ASM "\n";
    for ($i = 0; $i < $index; ++$i) {
        print ASM "#line $line_num \"$proj_file\"\n";
        print ASM ".extern $cgi_map[$i]\n";
        print ASM "#line $line_num \"$proj_file\"\n";
        printf(ASM  "    pchkfile \"/%s.cgi\",%s\n", $cgi_map[$i], "$cgi_map[$i]");
    }
    print ASM "\n";
##    for ($i = 0; $i < $index; ++$i) {
##        print ASM "#line $line_num \"$proj_file\"\n";
##        printf(ASM  "    pchkfile \"/iu%02x\",%s\n", $i, "$cgi_map[$i]");
##    }
##    print ASM "\n";
    print ASM "#ifndef NO_PCODE\n";
    print ASM ".global pfindfile_end\n";
    print ASM "pfindfile_end:\n";

    print ASM "#line $line_num \"$proj_file\"\n";
    print ASM "    pprint \"HTTP/1.1 404 Not Found\\n\\n\"\n";

    print ASM "    pret\n";
    print ASM "    .word 0\n\n";
    print ASM "#endif /*NO_PCODE*/\n";
    print ASM "\n.section seeprom_cgi_strings\n";
}

sub sprintsn {
    local($s, $n) = @_;
    local($s1, $len);

    $s1 = substr($s, 0, $n);
    $s1 =~ s/[\r\n]/|/g;
    $len = length($s);
    ##return sprint "\"$s1\",len=$len"; 
    return "\"$s1\",len=$len"; 
}

sub process_html_tags {
    local($file, $line_num, $str) = @_;
    local($key, $len, $i, $s1, $s2, $s3, $s4, $s5, $s6);
    local($l1, $l2, $else_routine, $if_routine, $if_label);
    local($goto, $not);

    $key = $file; $key =~ s/[.]/_/g;

    if ($debug) {
           print "process_html_tags($file, $line_num, ".&sprintsn($str,30)."\n";
    }
    $str =~ s/[\n]/\r/g ; # map new-lines to CR

    # process `?...{}
    while (($i = index($str, '`?')) >= 0) {
        if ($i > 0) {
            $s1 = substr($str,0,$i);
            &new_string(1, $file, $line_num, $s1);
            $str = substr($str,$i); $line_num += &count_lines($s1);
        }
        $goto = '';
        $not = '';
        # process `?xxx.cgi?hh`<pw_if>...<pw_else>...<pw_endif>
        # process `?!xxx.cgi?hh`<pw_if>...<pw_else>...<pw_endif>
        if ($str =~ /^`[?]([!]*)([a-z0-9_]+)\.cgi[?]([^`\r]*)`\s*\<pw_if\>(.*)\<pw_else\>(.*)\<pw_endif\>(.*)$/i) {
           $not = $1;
           $s1 = "psetparm $3";
           $s2 = "pcall $2";
           $s3 = $4;
           $s4 = $5;
           $s5 = $6;
           if ($debug) {
               print "-------------------------------------------------\n";
               print "`?$1?$2`\n";
               print "<pw_if>\n";
               print &sprintsn($s3,40);
               print "\n<pw_else>\n";
               print &sprintsn($s4,40);
               print "\n<pw_endif>\n";
               print "-------------------------------------------------\n";
           }
        # process `?xxx.cgi?hh@yyy.cgi`
        # process `?!xxx.cgi?hh@yyy.cgi`
        } elsif ($str =~ /^`[?]([!]*)([a-z0-9_]+)\.cgi[?]([^`\r@]*)\s*[@]\s*([a-z0-9_]+).cgi`(.*)$/i) {
           $not = $1;
           $s1 = "psetparm $3";
           $s2 = "pcall $2";
           $s3 = '';
           $s4 = '';
           $s5 = $5;
           $goto = $4;
           if ($debug) {
               print "-------------------------------------------------\n";
               print "`?$1?$2`\n";
               print &sprintsn($s3,40);
               print "\n{\n";
               print &sprintsn($s4,40);
               print "\n}\n";
               print "-------------------------------------------------\n";
           }
        # process `?xxx.cgi?hh`...{}
        # process `?!xxx.cgi?hh`...{}
        } elsif ($str =~ /^`[?]([!]*)([a-z0-9_]+)\.cgi[?]([^`\r]*)`([^{]*)[{]([^}]*)[}](.*)$/i) {
           $not = $1;
           $s1 = "psetparm $3";
           $s2 = "pcall $2";
           $s3 = $4;
           $s4 = $5;
           $s5 = $6;
           if ($debug) {
               print "-------------------------------------------------\n";
               print "`?$1?$2`\n";
               print &sprintsn($s3,40);
               print "\n{\n";
               print &sprintsn($s4,40);
               print "\n}\n";
               print "-------------------------------------------------\n";
           }
         # process `?hh{} [old way!]
         } elsif ($str =~ /^`[?]([0-9a-f][[0-9a-f])([^{]*)[{]([^}]*)[}](.*)$/i) {
           ###$str = $1 . '0' . $2;   # map `?hh --> `0hh
           $s1 = "psetparm 0x$1";
           $s2 = "pcall testport";
           $s3 = $2;
           $s4 = $3;
           $s5 = $4;
           if ($debug) {
              print "-------------------------------------------------\n";
              print "?$1\n";
              print &sprintsn($s3,40);
              print "\n{\n";
              print &sprintsn($s4,40);
              print "\n}\n";
              print "-------------------------------------------------\n";
           }
        } else {
           $s1 = substr($str, 20); $s1 =~ s/[\r].*//;
           print "$file($line_num): badly formated `? expression ($s1...)\n";
           exit 1;
        }

        if ($s3 eq '') {
            $if_routine = '0';
        } else {
            $if_routine = "__if_$key"."_$numstrings"."_$line_num";
        }
        if ($s4 eq '') {
            $else_routine = '0';
        } else {
            $else_routine = "__else_$key"."_$numstrings"."_$line_num";
        }
        $if_label = "__if_jump_$key"."_$numstrings"."_$line_num";

        $str = $s5;

        if ($s1 ne '') {
            $pcode{$file} .= "#line $line_num \"$file\"\n";
            $pcode{$file} .= "    $s1\n" if ($s1 ne '');
        }
        $pcode{$file} .= "#line $line_num \"$file\"\n";
        $pcode{$file} .= "    $s2\n";

        if ($goto ne '') {
            $pcode{$file} .= "#line $line_num \"$file\"\n";
            if ($not eq '') {
                $pcode{$file} .= "    pjumpeq $goto\n";
            } else {
                $pcode{$file} .= "    pjumpne $goto\n";
            }
        } else {
            $pcode{$file} .= "#line $line_num \"$file\"\n";
            if ($not eq '') {
                $pcode{$file} .= "    pcifelse $if_routine,$else_routine\n";
            } else {
                $pcode{$file} .= "    pcifelse $else_routine,$if_routine\n";
            }
            $pcode{$file} .= "#line $line_num \"$file\"\n";
            $pcode{$file} .= "    pjump $if_label\n";
            if ($if_routine ne '0') {
                $pcode{$file} .= "$if_routine:\n";
                $pcode{$file} .= &new_string(0, $file, $line_num, $s3);
                $line_num += &count_lines($s3);
                $pcode{$file} .= "    pret\n\n";
            }
            if ($else_routine ne '0') {
                $pcode{$file} .= "$else_routine:\n";
                $pcode{$file} .=  &new_string(0, $file, $line_num, $s4);
                $line_num += &count_lines($s4);
                $pcode{$file} .= "    pret\n\n";
            }
            $pcode{$file} .= "#line $line_num \"$file\"\n";
            $pcode{$file} .= "$if_label:\n";
        }
    }

    # process CGI tags (`xxx.cgi?parm`)
    while ($str =~ /^([^`]*)`([_a-z0-9]+)\.cgi[?]([^`\r]*)`(.*)$/i) {
       $s1 = $1;
       $s2 = $2;
       $s3 = $3;
       $s4 = $4;
       &new_string(1, $file, $line_num, $s1);
       $str = $s4; $line_num += &count_lines($s1);
       if ($debug) {
           $l1 = length($s1);
           $l3 = length($s4);
           print "---split $file: on $s2?$s3 ($l1/$l3)\n";
       }
 $line_num = 0 if ($line_num < 0);
       $pcode{$file} .= "#line $line_num \"$file\"\n";
       if ($s3 =~ /^\s*(\d+)\s*$/) {
           $pcode{$file} .= "    psetparm $1\n";
       } elsif ($s3 =~ /^\s*(0x[0-9a-fA-F]+)\s*$/) {
           $pcode{$file} .= "    psetparm $1\n";
       } elsif ($s3 =~ /^\s*["](.*)["]\s*$/) {
           $pcode{$file} .= "    psetparm \"$1\"\n";
       } else {
           print "$file($line_num): badly formated parameter in `?routine.cgi?parm` expression ($s3)\n";
           $pcode{$file} .= "    psetparm 0\n";
           exit 1;
       }
       $pcode{$file} .= "#line $line_num \"$file\"\n";
       $pcode{$file} .= "    pcall $s2\n";
    }

    # process CGI tags (`xxx.cgi`)
    while ($str =~ /^([^`]*)[`]([_a-z0-9][_a-z0-9]*)\.cgi[`](.*)$/i) {
       $s1 = $1;### $s1 =~ s/[\r]/\n/g ;
       $s2 = $2;
       $s3 = $3;
       &new_string(1, $file, $line_num, $s1);
       $str = $s3; $line_num += &count_lines($s1);
       if ($debug) {
           $l1 = length($s1);
           $l3 = length($s3);
           print "---split $file: on $s2 ($l1/$l3)\n";
       }
       $pcode{$file} .= "#line $line_num \"$file\"\n";
       $pcode{$file} .= "    pcall $s2\n";
    }

    # process CGI tags (`@xxx.cgi`)
    while ($str =~ /^([^`]*)[`][@]([_a-z0-9][_a-z0-9]*)\.cgi[`](.*)$/i) {
       $s1 = $1;### $s1 =~ s/[\r]/\n/g ;
       $s2 = $2;
       $s3 = $3;
       &new_string(1, $file, $line_num, $s1);
       $str = $s3; $line_num += &count_lines($s1);
       if ($debug) {
           $l1 = length($s1);
           $l3 = length($s3);
           print "---split $file: on goto (@) $s2 ($l1/$l3)\n";
       }
       $pcode{$file} .= "#line $line_num \"$file\"\n";
       $pcode{$file} .= "    pjump $s2\n";
    }

    # process "label CGI tags (`=xxx.cgi`)
    while ($str =~ /^([^`]*)[`][=]([_a-z0-9][_a-z0-9]*)[`](.*)$/i) {
       $s1 = $1;### $s1 =~ s/[\r]/\n/g ;
       $s2 = $2;
       $s3 = $3;
       &new_string(1, $file, $line_num, $s1);
       $str = $s3; $line_num += &count_lines($s1);
       if ($debug) {
           $l1 = length($s1);
           $l3 = length($s3);
           print "---split $file: on label (=) $s2 ($l1/$l3)\n";
       }
       $pcode{$file} .= "#line $line_num \"$file\"\n";
       $pcode{$file} .= $s2."::\n";
    }

    $str =~ s/[\r]/\n/g ; # put back new-lines
    return ($str, $line_num);
}

sub count_lines {
    local($s) = @_;

    $s =~ s/[^\r]//g;
    ###print STDERR length($s)."\n";
    return length($s);
}

sub print_file_def_line {
   local($file) = @_;
   local($i);

   $i = $file_def_line{$file};
   if ($i < 1) {
       ###print "?? $file\n";
       $i = 9999;
   }
   print ASM "#line $i \"$proj_file\"\n";
}

