//-------------------------------------------------------------------------
// Copyright (c) 1999 Steven Freyder, David Helland, Bruce D. Lightner and
// Lightner Engineering.  Use subject to licensing restrictions described
// the file LICENSE.txt.  If you do not agree with the licensing terms then
// you must return and/or destroy this software immediately!
//-------------------------------------------------------------------------

/************************************************************************/
/*
 * this is the canonical command port used by the various batch files
 * which invoke the debugger via the 'net.  if you don't have DEBUGGER
 * defined, then it doesn't matter because there won't be any command
 * interpreter to run the commands anyway.  Unless of course your code
 * has some special conditionals which don't depend on the DEBUGGER
 * symbol...
 */
#ifndef TCPPORT_CMD
#define TCPPORT_CMD 911
#endif

#ifndef SER_EEPROM_BASE_ADDR
#define SER_EEPROM_BASE_ADDR    0xA0	/* 24LC128 128K-bit serial EEPROM */
#endif

/*
 * define this if you don't want printout of your ether address on
 * boot.
 */
//#define NO_ETHER_ADDR_ON_BOOT

/*
 * idle network packet watchdog timer limit (range: 100-254 loop counts)
 */
#ifndef PKT_WATCHDOG_MAX
#define PKT_WATCHDOG_MAX    250
#endif

/*
 * default bootup banner sent to serial port
 */
#ifndef BANNER
#ifdef BREADBOARD
#define BANNER "\r\nPicoWeb Breadboard\r\n"
#else
#define BANNER "\r\nPicoWeb Server\r\n"
#endif
#endif

/*
 * default microcontroller clock rate
 */
#ifndef CLOCK
#define CLOCK 7372000
#endif

/*
 * default serial port baud rate clock divisor
 *
 *               - Processor  -
 *   Baud Rate   - Clock Rate -
 *    Divisor    8 MHz    4 MHz
 *    -------    -----    -----
 *       25      19200     9600
 *       51       9600     4800
 *      103       4800     2400
 *      207       2400     1200
 */

#ifndef BAUD_RATE
#define BAUD_RATE 19200
#endif

#ifndef SERIAL_BAUD_DIVISOR
#define UBRR(c, b)   (((((c) * 10) / ((b) * 16)) + 5) / 10 - 1)
#define SERIAL_BAUD_DIVISOR UBRR(CLOCK, BAUD_RATE)
#endif

/*
 * set marco for serial port pcode routine "psgetcto"
 */
#include "psgetcto.h"

/*
 * default to store IP address in EEPROM if NET_CONFIG_IP enabled
 */
#ifdef NET_CONFIG_IP  /* allow IP address reconfiguration via network? */
# ifndef EEPROM_IP
#  ifndef SEEPROM_IP
#   define EEPROM_IP
#  endif
# endif
#endif

/************************************************************************/
