#include "defdefs.h"

.global baud_rate_div100
.equ baud_rate_div100,(BAUD_RATE/100)

.global clock
.equ clock,CLOCK

;------------------------------------------------------------------------
.section reset_00

#ifdef DEBUGGER
#ifdef BANNER
    pbegin
    pprint BANNER
    pend
#endif
#endif

#ifdef ENABLE_WATCHDOG
    ldi r16,0x0F    ; enable watchdog timer (~2 sec.)
    out wdtcr,r16
    wdr             ; reset watchdog timer
#endif

;------------------------------------------------------------------------
.section reset_end

#ifdef DEBUGGER
#ifndef NO_ETHER_ADDR_ON_BOOT
    rcall show_ether
    rcall do_crlf
#endif
#endif
;------------------------------------------------------------------------
.section fast

#ifdef DEBUGGER
    clr r0
    sts putc_b,r0
    rcall command_check
#endif

#ifdef ENABLE_WATCHDOG
    wdr             ; reset watchdog timer
#endif
;------------------------------------------------------------------------
.section slow

#if defined(USE_BOOTP)
    rcall bootp_timer
#endif

;------------------------------------------------------------------------
.text
;+
; **-pwdr-pcode watchdog reset.
;
; inputs:
;       none
;-
    .global pwdr
    .global swdr
pwdr:   pcode_routine   0
swdr:
#ifdef ENABLE_WATCHDOG
    wdr             ; reset watchdog timer
#endif
    ret
;------------------------------------------------------------------------
.text
#ifndef NO_PCODE
#ifndef DEBUGGER
    .global tcpget_cmdline
tcpget_cmdline:
    pret
#endif
#endif
;------------------------------------------------------------------------
.section eeprom
#ifndef BREADBOARD
    .global eeprom_ether
eeprom_ether:
#include "ether.inc"
#endif

    .global baud_rate_divisor
baud_rate_divisor:
    .byte SERIAL_BAUD_DIVISOR

;------------------------------------------------------------------------
.text
#ifndef NO_PCODE
.global unknown_tcp_port
unknown_tcp_port:
#ifdef UNKNOWN_TCP_PORT_HANDLER
    pjump UNKNOWN_TCP_PORT_HANDLER
#else
    pret
#endif
#endif

.global poksetip
#ifndef POKSETIP
poksetip: pcode_routine 0
    clr r16                ; set Z-flag (i.e., SETIP changes allowed)
    rjmp set_pcode_flags   ; set pcode flags
#else
.equ poksetip,POKSETIP
#endif

;------------------------------------------------------------------------
#ifdef SERIAL_FULL_PARITY

#ifndef SER_PARITY
#define SER_PARITY 2
#endif

#ifndef SER_STOP
#define SER_STOP 1
#endif

#ifndef SER_PAR_ERR
#define SER_PARITY 0
#endif

.section eeprom
.global baud_rate
baud_rate:   .word baud_rate_div100  ; baud rate / 100
.global parity
parity:      .byte SER_PARITY        ; 0=even, 1=odd, 2=none
.global stop_bits
stop_bits:   .byte SER_STOP          ; 1 or 2
.global parity_err
parity_err:  .byte SER_PAR_ERR       ; 0=OK, 1=drop, 2=set MSB

#endif /* SERIAL_FULL_PARITY */
;------------------------------------------------------------------------
#ifdef RESERVED_PAGES
.pushsection ne2000_resv
.global sm_reserved
sm_reserved:
	.skip RESERVED_PAGES<<8
.popsection

#endif /* RESERVED_PAGES */
;------------------------------------------------------------------------
;
; set size of global scrathc buffer "buf"
;
#ifndef NO_PCODE
#ifndef TMP_BUF_LEN
#define TMP_BUF_LEN 48      /* default bytes in global buffer "buf" */
#endif

#ifndef MIN_TMP_BUF_LEN
#define MIN_TMP_BUF_LEN 16  /* minimum size of global buffer "buf" */
#endif

.global buf_len
.equ buf_len,TMP_BUF_LEN

.pushsection .bss
.global buf
buf: .skip MIN_TMP_BUF_LEN              ; global scratch buffer
     .skip TMP_BUF_LEN-MIN_TMP_BUF_LEN
.popsection
#endif /* !NO_PCODE */
;------------------------------------------------------------------------
;
;  set size of serial port input queue (see serial.asm)
;
;  !!! If NOT power of 2, do not #define POWER_OF_2_QUEUE in serial.asm!!
;
#ifndef NO_SERIAL
#ifndef SER_INBUF_LEN
#define SER_INBUF_LEN 64
#endif

.global max_nch
.global MAX_NCH
.equ  max_nch = SER_INBUF_LEN  ; size of serial input queue

;
; serial port input queue (see serial.asm)
;
.pushsection .bss
.global inbuf
inbuf: .skip max_nch   ; SRAM serial port input queue
.popsection
#endif /* !NO_SERIAL */
;------------------------------------------------------------------------
#ifndef NO_DO_PUTCHAR
;+
; **-do_putchar-put character to appropriate place.
;
; inputs:
;   ch  = character (r24)
;   putc_b = controls destination:
;               0: output to UART
;            0xde: rcall import_user_debug (if #define USER_MODE_DEBUG)
;             <>0: output to Ethernet controller
;
; outputs:
;   character dispatched to: serial port, network, or nothing.
;
; note: callable by avr-gcc routines (preserves needed registers)
;
;-
.text
.global do_putchar
do_putchar:

#ifdef HOOK_PUTCHAR
    HOOK_PUTCHAR
#endif /*HOOK_PUTCHAR*/

#ifdef NO_PCODE
    rcall import_serial_putchar   ; "tiny kernel" option
    ret
#else /*!NO_PCODE*/
    push    r17
    lds r17,putc_b                ; check if going to board
    tst r17
    brne putc_to_board
    rcall import_serial_putchar   ; output character to UART
putc_off:
    pop r17
    ret
putc_to_board:
#if defined(USER_MODE_DEBUG)
    cpi r17,0xde        ; check if user mode DEbug
    breq go_user_debug
#endif /*USER_MODE_DEBUG*/
    rcall do_bputc                 ; output character to network
    rjmp putc_off
#if defined(USER_MODE_DEBUG)
go_user_debug:
    rcall import_user_debug
    rjmp putc_off
#endif /*USER_MODE_DEBUG*/
#endif /*!NO_PCODE*/
#endif /*!NO_DO_PUTCHAR*/
;------------------------------------------------------------------------
