;-------------------------------------------------------------------------
; Copyright (c) 1999 Steven Freyder, David Helland, Bruce D. Lightner and
; Lightner Engineering.  Use subject to licensing restrictions described
; the file LICENSE.txt.  If you do not agree with the licensing terms then
; you must return and/or destroy this software immediately!
;-------------------------------------------------------------------------

.listmac

#include "userdef.h"
#include "defdefs.h"
#include "picbin.inc"
;
; local data.
;
; this consists of data needed by this module, or anything that might be
; considered "global".  individual packages should have their data
; declarations local to themselves, so that if they are not included,
; there will be no hit on the data space.
;
.dseg
timer:          .byte   4       ; loop and timer counter
pkt_wdog_cnt:   .byte   1       ; packet watchdog counter

.section eeprom
tcp_hdr_debug_flag:
    .dw 0xff   ; dump TCP headers if not 0xff
.cseg

;
; default interrupt handler for unused interrupt vector table entries
;
default_isr:
    reti    ; return from interrupt
;
; this is the processor reset routine.  whenever something bad happens, or
; we get a real reset, control is transferred here.  remember that you can't
; do any rcall instructions until the stack is set up.
;
reset:
    clr r17
    out sreg,r17
    out gimsk,r17           ; disable INT0/INT1
    out mcucr,r17
#ifdef PKT_WATCHDOG_DEBUG
    sts pkt_wdogs,r17       ; zero packet watchdog debug counter
#endif

;
; when the DEBUGGER conditional is defined, pressing Control-C will transfer
; control here, assuming that the user's SERIAL.ASM module facilitates it.
;
CONTROL_C:
    watchdog_init                   ; initialize the watchdogs  **PICOWEB
    stack_init                      ; reset the stack pointer   **PICOWEB

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;; DO NOT DO ANY [R]CALL instructions before this point! ;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#ifndef USER_rx_isr
    rcall serial_init
#endif
    ;;;rcall   i2c_init      ; done automatically now!!

    sei
    picoweb_init                    ;                   **PICOWEB

    rcall getip
;
; application-specified reset code will be included here
;
#ifdef USER_RESET_ASM
#include USER_RESET_ASM
#endif

#ifdef DEBUGGER
#ifdef BANNER
    outstr BANNER      ; output bootup banner to serial port
#endif
#ifndef NO_ETHER_ADDR_ON_BOOT
    rcall show_ether
#endif
    ldsw yl,ccaddr                  ; show PC (from ^C interrupt)
    crlf
    putchar '@'
    rcall hexy
#endif

doprompt:

#if defined(DEBUGGER)
    rcall do_prompt
    clr r17
    sts     nch,r17
    sts     ready,r17
    inc r17
    sts neon,r17
    sts nelog,r17
#else
    clr r0
    sts nelog,r0                ; no debugger means no output from net
#endif

;
; this counter wraps about every 1.5 seconds on an 8515/4 with the current
; processing load.  SDF 1998/11/15.
;
idle_loop:
    clr r0
    ldi r17,1
    lds r16,timer
    add r16,r17
    sts timer,r16
    lds r16,timer+1
    adc r16,r0
    sts timer+1,r16
    brcc fast_idle
;
; the "slow idle" timer has counted down.
;
slow_idle:
#if defined(USE_BOOTP)
    rcall bootp_timer
#endif
    network_watchdog        ;                           **PICOWEB
;
; application-specified "slow idle" code will be included here
;
#ifdef USER_SLOW_IDLE_ASM
#include USER_SLOW_IDLE_ASM
#endif
.cseg

;
; slow idle loop - wait for CR, call network handler.
;
fast_idle:

#ifdef ENABLE_WATCHDOG
    wdr             ; reset watchdog timer
#endif
;
; check if network code is enabled, if so, set logging flag based on nelog, and
; call network logic, then enable logging.
;
#if defined(DEBUGGER)
    lds r17,neon
    or r17,r17
    breq net_off
#endif

    lds r17,nelog
    sts putcok,r17          ; enable/disable output based on nelog flag
    rcall network           ; propagate network I/O     **PICOWEB
    ser r17
    sts putcok,r17          ; turn output back on
#if defined(DEBUGGER)

net_off:
    rcall command_check
#endif

;
; application-specified "fast idle" code will be included here
;
#ifdef USER_FAST_IDLE_ASM
#include USER_FAST_IDLE_ASM
#endif
.cseg

    rjmp idle_loop

;+
; **-pwdr-pcode watchdog reset.
;
; inputs:
;       none
;-

pwdr:   pcode_routine   0
#ifdef ENABLE_WATCHDOG
    wdr             ; reset watchdog timer
#endif
    ret

.cseg
#include "serial.asm"
#include "getip.asm"

#if defined(DEBUGGER)
#include "cmd.asm"
#endif

#if !defined(NO_TCP)
#include "usertcp.asm"
#endif

#include "i2cprog.asm"
#include "seeload.asm"

;
; application-specified code will be included here
;
#ifdef USER_ASM
#include USER_ASM
#endif
.cseg

;
; include the strings which have been created by ppre.pl.  note that this
; MUST be a .include and not a #include!!!!
;
.cseg
#ifdef PW_STRINGS
.include PW_STRINGS
#else
.include "picweb.asi"
#endif
