;--------------------------------------------------------------------------
;+
; **-pprintnswi-output "narrow" signed 16-bit integer in decimal (no spaces)
;
; pprintnswi value
;
; inputs:
;   value = 16-bit word to convert
;
; outputs:
;   result printed (with pputc), right-justified in 6-character field
;
; working SRAM:
;   15 bytes (save/restores on stack)
;
;--------------------------------------------------------------------------
#define T1 (BP+0)
#define TEMPS_END (T1+5)
#define N_TEMPS (TEMPS_END-T1)

#define P1    TEMPS_END
#define P2    P1+2
#define P3    P2+2

#define DIG_BUF T1        /* uses 5 bytes for ASCII digit storage */
#define DIG_PTR P3
#define DEC_REM P2        /* must be adjacent in RAM (high) */
#define DEC_VAL P1        /* must be adjacent in RAM (low) */

.text
.global pprintnswi
pprintnswi: pcode_routine  1
    rcall pip_setup
    pjump printint

.global pprintnuwi
pprintnuwi: pcode_routine  1
    rcall pip_setup
    pjump printint_unsigned

#if 0
.section seeprom
#else
.text
#endif

printint_unsigned:
    penter N_TEMPS
    pjump printint0

printint:
    penter N_TEMPS
    pbitwi DEC_VAL,0x8000               ; check the sign
    pjumpeq printint0                   ; positive - start conversion
    pnegw DEC_VAL                       ; negate it
    pputc '-'

printint0:
    pmovwi DIG_PTR,DIG_BUF              ; setup output buffer pointer
printint1:
    pdiv DEC_VAL,[DEC_VAL],10           ; get next one
    paddwi DEC_REM,'0'                  ; convert digit for display
    pmovbi [DIG_PTR],[byte DEC_REM]     ; save digit
    paddwi DIG_PTR,1                    ; bump past digit just stored
    psubwi DEC_VAL,0                    ; test remaining value
    pjumpne printint1                   ; back if more to convert
printint2:
    psubwi DIG_PTR,1                    ; point to next one
    pputcb [DIG_PTR]                    ; show it
    pcmpwi DIG_PTR,DIG_BUF              ; check if just did last one
    pjumpne printint2                   ; nope - do next one
    pretn N_TEMPS                       ; fix stack and return to caller

