#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bootp.h"
#include "getopt.h"

/************************************************************************/

#ifdef SunOS
#define UNIX
#endif

#if defined(UNIX)
#include    <unistd.h>
#include    <errno.h>
#include    <sys/types.h>
#include    <sys/stat.h>

#ifdef SunOS
#include    <malloc.h>
#include    <sys/filio.h>
#else /* !SunOS */
#include    <sys/select.h>
#include    <asm/ioctls.h>
#endif /* !SunOS */

#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#define SockClose(x)  close(x)

#ifndef O_BINARY
#define O_BINARY 0
#endif

#endif /* UNIX */

#if defined(_CONSOLE)
#define IS_WINDOWS
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <winsock.h>

#define SockClose(x)    closesocket(x)
#define sleep(x)        Sleep((x)*1000L)
#define NetErrno() (WSAGetLastError())

#else /* !_CONSOLE */

#define NetErrno() ((long)errno)
#define GetLastError() ((long)errno)
#define WSADATA int
int WSAStartup(x, y) { return 0; }
#define MAKEWORD(x,y)  (((x)<<8) | (y))

void strlwr(char *s) {
    int ch;
    while (*s) {
        *s = tolower(*s);
	++s;
    }
}

#endif /* !_CONSOLE */

#ifndef INADDR_NONE
#define INADDR_NONE (-1)
#endif

#if !defined(TRUE)
#define TRUE (0==0)
#endif

#if !defined(FALSE)
#define FALSE (0!=0)
#endif

#if !defined(WORD)
#define WORD unsigned short
#endif

#if !defined(BOOLEAN)
#define BOOLEAN int
#endif

/************************************************************************/

#define BOOTP_PORT 67
#define     FMT_SIZE    16      /* format size */

static BOOLEAN bDebug = FALSE ;
static BOOLEAN bDump = FALSE ;

/*
 * INPUTS:
 *  buffer      = pointer to input buffer
 *  n_bytes     = number of bytes to dump in the buffer
 *  record      = record number for header (0 = no header)
 *  output      = file pointer to receive resulting output
 *
 * OUTPUTS:
 *
 * RETURNS:
 *
 * CALLS:
 *
 * DESCRIPTION:
 *
 */
dumper(buffer, n_bytes, record, output)

    char    *buffer ;           /* */
    int n_bytes ;           /* number of bytes to dump */
    FILE    *output ;           /* pointer to output */

{

    register int i, j ;         /* loop counters */
    int offset_count ;          /* offset counter */
    int working_size ;          /* current working size */
    int c ;             /* a character */
    char    *ch_ptr ;           /* a character pointer */
    char    hex_buf[(FMT_SIZE*3)+1] ;   /* hex buffer */
    char    asc_buf[FMT_SIZE+1] ;       /* ascii buffer */
    static  char hex_digits[] = { "0123456789ABCDEF" } ;


    if (record > 0)  {          /* if want header */
      fprintf(
            output,
            "Record number %d (%08X), %d (%04X) byte%s\n\n",
            record,record,n_bytes,n_bytes,
            (n_bytes!=1 ? "s" : "")
            ) ;
    }

    offset_count = 0 ;          /* reset offset counter */
    hex_buf[(FMT_SIZE*3)] = '\0' ;      /* terminate hex buffer */
    asc_buf[FMT_SIZE] = '\0' ;      /*   and ascii buffer */

    for (i=0; n_bytes > 0; ) {      /* for entire buffer */
      if (n_bytes >= FMT_SIZE)  {   /* if number of remaining bytes GE */
        working_size = FMT_SIZE ;       /* format size then */
      }                 /*   set working size to max */
      else {                /* otherwise... */
        working_size = n_bytes ;        /*   just keep the number */
      }
      ch_ptr = &hex_buf[3*FMT_SIZE-1] ; /* pointer to hex buffer */
      for(j=0 ; j < working_size ; j++)  {  /* for each byte */
        c = buffer[i++]&0xff ;      /* get next byte */
        if ((c < 32) || (c>=127)) {     /* if c is a non-printable */
          asc_buf[j] = '.' ;        /*  subsitute a dot for it */
        }
        else {              /* otherwise... */
          asc_buf[j] = c ;          /* just keep the character */
        }
        *ch_ptr-- = hex_digits[c&0xf] ; /* bottom */
        *ch_ptr-- = hex_digits[(c>>4)&0xf] ; /* and top */
        *ch_ptr-- = ' ' ;           /*   and a blank */
      }

      while (ch_ptr >= hex_buf)  {      /* while not start of buffer */
        *ch_ptr-- = ' ' ;           /* make it a blank */
      }

      for (j=working_size ; j<FMT_SIZE ; j++)  {
        asc_buf[j] = '.' ;          /* blank out unused ascii */
      }

      fprintf(output, "%s %s %06X\n", hex_buf, asc_buf, offset_count) ;
      offset_count = offset_count + working_size ;
      n_bytes = n_bytes - working_size ;
    }

    return(1) ;
}

static void
Err(
    char    *why
){
#ifdef IS_WINDOWS
    LPVOID lpMsgBuf = NULL;
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
              FORMAT_MESSAGE_FROM_SYSTEM |
              FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        WSAGetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
        (LPTSTR) &lpMsgBuf,
        0,
        NULL );
    if (lpMsgBuf == NULL) lpMsgBuf = "???\n";
    fflush(stdout);
    fprintf(stderr,"%s: Error %lu, %lu: %s",
        why,GetLastError(),WSAGetLastError(),lpMsgBuf) ;
    LocalFree(lpMsgBuf) ; // free the buffer
#else
    fflush(stdout);
    perror(why);
#endif
    exit(2) ;
}

static int
UdpBroadcast(
    int Socket,
    void *Buf,
    int Len,
    struct sockaddr_in *pToSocket
){
    struct sockaddr_in SendSockAddr;
    int len = sizeof(SendSockAddr);

    if (getsockname(Socket,(struct sockaddr *)&SendSockAddr,&len) < 0 || len != sizeof(SendSockAddr)) {
        Err("getsockname");
    }
#if 1
    pToSocket->sin_addr.s_addr = INADDR_BROADCAST ;
#else
    pToSocket->sin_addr.s_addr = 0xffe3f5c0 ;
#endif
    pToSocket->sin_port = htons(BOOTP_PORT+1) ;

 /////pToSocket->sin_addr.s_addr = SendSockAddr.sin_addr.s_addr ;
 /////pToSocket->sin_port = SendSockAddr.sin_port ;

    ////SendSockAddr.sin_port = htons(BOOTP_PORT+1) ;

    if (bDebug || bDump)  {
        //fprintf(stdout,"\nUdpBroadcast to %lx:\n",(unsigned long)(pToSocket->sin_addr.s_addr));
        fprintf(stdout,"\nUdpBroadcast to %lx:%d ",
                (unsigned long)(pToSocket->sin_addr.s_addr),
                (int)ntohs(pToSocket->sin_port));
        fprintf(stdout,"using socket %lx:%d\n",
                (unsigned long)(SendSockAddr.sin_addr.s_addr),
                (int)ntohs(SendSockAddr.sin_port));
        if (bDump) dumper(Buf,Len,0L,stdout) ;
    }
    if (sendto(Socket,Buf,Len,0,(struct sockaddr *)pToSocket,sizeof *pToSocket) == Len)  {
        return(0) ;
    }
    return(GetLastError()) ;
}

void Usage(char *program) {
    fprintf(stderr,
            "usage: %s [options]\n"
            "where options:\n"
            "  -d           - set debug mode\n"
            "  -i IPaddr    - bind to IP address \"IPaddr\"\n"
            "  -p port      - bind to port \"port\" (default %d)\n"
            "  -f file      - use file \"file\" for BOOTP database\n"
            "  -n           - don't respond to BOOTP requests (debug)\n"
            "  -B IPaddr    - send repeated BOOTP requests to IP address \"IPaddr\" (debug)\n"
            "  -N num       - number of repeats for \"-B\" requests (debug)\n"
            "  -x           - dump packets in hex (debug)\n"
        ,program,BOOTP_PORT) ;
}

main(
    int argc,
    char *argv[]
){
    int         ListenSocket ;
    int         ReplySocket ;
    WORD        wVer ;
    WSADATA     wsaData ;
    struct      sockaddr_in SockAddr ;
    struct      sockaddr_in FromAddr ;
    struct      sockaddr_in ReplyAddr ;
    int         iAddrLen ;
    int         n = 10 ;
    int         port = -1 ;
    char        *ip = NULL ;
    char        *s ;
    FILE        *fp ;
    char        FileEth[128] ;
    char        FileIp[128] ;
    char        FileVend[128] ;
    char        Ether[32] ;
    char        Rec[128] ;
    char        BootpBuf[1460] ;
#define Bootp (*((BOOTP_PACKET *)BootpBuf))
    char        c ;
    char        *BootpFile = "bootp.dat" ;
    BOOLEAN     bRespond = TRUE ;
    char        *sBangAddr = NULL ;
    char        *program ;

    program = argv[0];
#define CMD_SWITCHES "dxnf:i:p:B:N:?h"
    while ((c = getopt(argc,argv,CMD_SWITCHES)) != EOF)  {
        switch (c)  {
            case 'd':  {
                bDebug = TRUE ;
                break ;
            }
            case 'x':  {
                bDump = TRUE ;
                break ;
            }
            case 'n':  {
                bRespond = FALSE ;
                break ;
            }
            case 'p':  {
                port = atoi(optarg) ;
                break ;
            }
            case 'i':  {
                ip = optarg ;
                break ;
            }
            case 'f':  {
                BootpFile = optarg ;
                break ;
            }
            case 'B':  {
                sBangAddr = optarg ;
                break ;
            }
            case 'N':  {
                n = atoi(optarg) ;
                break ;
            }
            case '?':
                Usage(program);
                return(0);
            default:  {
                Usage(program);
                return(1);
            }
        }
    }

    wVer = MAKEWORD(1,1) ;              // request WinSock version 1.1
    if (WSAStartup(wVer,&wsaData) != 0)  {  // if startup failed
        Err("wsastartup") ;
    }

    ListenSocket = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP) ;
    if (ListenSocket < 0)  {
        Err("socket") ;
    }

    SockAddr.sin_family = AF_INET ;
    SockAddr.sin_port = htons(BOOTP_PORT) ;
#ifdef IS_WINDOWS
    SockAddr.sin_addr.s_addr = INADDR_ANY ;
#else
    SockAddr.sin_addr.s_addr = INADDR_BROADCAST ;
#endif

    if (port >= 0) {
        SockAddr.sin_port = htons((u_short)port) ;
    }

    if (bind(ListenSocket,(struct sockaddr *)&SockAddr, sizeof SockAddr) != 0)  {
        Err("bind") ;
    }

    memcpy(&ReplyAddr, &SockAddr, sizeof(ReplyAddr));
    if (ip) {
        ReplyAddr.sin_addr.s_addr = inet_addr(ip) ;
        ReplyAddr.sin_port = 0 ;  /* get any old port to reply on */
        if (ReplyAddr.sin_addr.s_addr == INADDR_NONE) {
            Err(ip);
        }
        ReplySocket = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP) ;
        if (ReplySocket < 0)  {
            Err("socket reply") ;
        }
        if (bind(ReplySocket,(struct sockaddr *)&ReplyAddr, sizeof ReplyAddr) != 0)  {
            Err("bind reply") ;
        }
        {
            BOOLEAN bBroad = TRUE ;
            if (setsockopt(ReplySocket,SOL_SOCKET,SO_BROADCAST,
                (char *)&bBroad,sizeof bBroad) != 0)  {
                Err("so_broadcast") ;
            }
        }
{
    struct sockaddr_in ReplySockAddr;
    int len = sizeof(ReplySockAddr);

    ReplyAddr.sin_addr.s_addr = INADDR_BROADCAST ;
    if (connect(ReplySocket,(struct sockaddr *)&ReplyAddr, sizeof ReplyAddr) != 0)  {
        Err("connect") ;
    }
    if (getpeername(ReplySocket,(struct sockaddr *)&ReplySockAddr,&len) < 0 || len != sizeof(ReplySockAddr)) {
        Err("getsockname");
    }
        fprintf(stdout,"peer socket %lx:%d\n",
                (unsigned long)(ReplySockAddr.sin_addr.s_addr),
                (int)ntohs(ReplySockAddr.sin_port));
}
    } else {
        ReplySocket = ListenSocket;
    }

    {
        BOOLEAN bBroad = TRUE ;
        if (setsockopt(ListenSocket,SOL_SOCKET,SO_BROADCAST,
            (char *)&bBroad,sizeof bBroad) != 0)  {
            Err("so_broadcast") ;
        }
    }

    if (sBangAddr != NULL)  {
        FromAddr.sin_family = AF_INET ;
        FromAddr.sin_port = htons(BOOTP_PORT+1) ;
        FromAddr.sin_addr.s_addr = inet_addr(sBangAddr) ;
        memset(&Bootp,0,sizeof Bootp) ;
        while (n-- > 0)  {
            if (UdpBroadcast(ReplySocket,&Bootp,10,&FromAddr) != 0)  {
                fflush(stdout);
                perror(" send error");
                fflush(stderr);
            }
            else  {
                printf("*") ;
            }
        }
        return(0) ;
    }
    for (;;)  {
        iAddrLen = sizeof FromAddr ;
        n = recvfrom(ListenSocket,(void *)&Bootp,sizeof BootpBuf,0,
            (struct sockaddr *)&FromAddr,&iAddrLen) ;
        if (bDebug || bDump)  {
            printf("OP %02X HW %02X HWL %02X %02X:%02X:%02X:%02X:%02X:%02X/%s:%d\n",
                Bootp.optype,Bootp.hwtype,Bootp.hwlen,
                Bootp.chaddr[0],
                Bootp.chaddr[1],
                Bootp.chaddr[2],
                Bootp.chaddr[3],
                Bootp.chaddr[4],
                Bootp.chaddr[5],
                inet_ntoa(FromAddr.sin_addr),
                ntohs(FromAddr.sin_port)) ;
            if (bDump) {
                dumper(&Bootp,n,0L,stdout) ;
                printf("\n") ;
            }
        }
        if (!bRespond)  {
            continue ;
        }
        if (ntohs(FromAddr.sin_port) == BOOTP_PORT)  {  /* bogus */
            if (bDebug) {
                fprintf(stderr, "Dropping bogus request!\n") ;
            }
            continue ;
        }
        fp = fopen(BootpFile,"r") ;
        if (fp != NULL)  {
            sprintf(Ether,"%02x:%02x:%02x:%02x:%02x:%02x",
                Bootp.chaddr[0],
                Bootp.chaddr[1],
                Bootp.chaddr[2],
                Bootp.chaddr[3],
                Bootp.chaddr[4],
                Bootp.chaddr[5]
            ) ;
            while (fgets(Rec,sizeof Rec-1,fp) != NULL)  {
                s = strchr(Rec,'\n') ;
                if (s != NULL)  {
                    *s = 0 ;
                }
                strlwr(Rec) ;
                sscanf(Rec,"%s %s %s",FileEth,FileIp,FileVend) ;
                if (strcmp(Ether,FileEth) == 0)  {
                    *((unsigned long *)&Bootp.yiaddr) = inet_addr(FileIp) ;
                    Bootp.optype = 2 ;      /* reply */
                    fprintf(stderr, "%s -> %s",FileEth,FileIp) ;
                    if (UdpBroadcast(ReplySocket,&Bootp,sizeof Bootp,&FromAddr) != 0)  {
                        fflush(stdout);
                        perror(" send error");
                        fflush(stderr);
                    }
                    else  {
                        fprintf(stderr,"\n") ;
                        fflush(stderr);
                    }
                    break ;
                }
            }
            fclose(fp) ;
        }
        else  {
            fprintf(stderr,"BOOTP: open failure on %s\n",BootpFile) ;
        }
    }
    return(0) ;
}
