#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <windows.h>
#include "regex.h"
#include "getopt.h"

static regex_t Pattern ;
static int bShowFileName = FALSE ;
static int bShowLineNumber = FALSE ;
static int bCountOnly = FALSE ;
static int bMultiFile = FALSE ;
static int bNotMatched = FALSE ;
static int iMatches = 0 ;
static int iFiles = 0 ;

static int
iGrepFile(
    char    *sFileName,
    FILE    *fp
){
    int     i ;
    char    sBuffer[1024] ;
    int     iLine = 0 ;
    char    *s ;

    iFiles++ ;
    while (fgets(sBuffer,sizeof sBuffer-1,fp) != NULL)  {
        if ((s=strchr(sBuffer,'\n')) != NULL)  {
            *s = 0 ;
        }
        iLine++ ;
        i = regexec(&Pattern,sBuffer,0,NULL,0) ;
        if (bNotMatched)  {
            i = (i==0 ? 1 : 0) ;
        }
        if (i == 0)  {          /* we matched */
            iMatches++ ;
            if (!bCountOnly)  {
                if (bShowFileName)  {
                    printf("%s:",sFileName) ;
                }
                if (bShowLineNumber)  {
                    printf("%d:",iLine) ;
                }
                printf("%s\n",sBuffer) ;
            }
        }
    }
    if (bCountOnly && !bMultiFile)  {
        if (bShowFileName)  {
            printf("%s:",sFileName) ;
        }
        printf("%d lines matched.\n",iMatches) ;
    }
    return(iMatches == 0 ? 1 : 0) ;
}

static void
vGrepWild(
    char    *sName
){
    WIN32_FIND_DATA FindData ;
    HANDLE  hFind ;
    FILE    *fp ;
    char    sFilePath[1024] ;
    char    *s ;
    int     iIndex = 0 ;

    s = strrchr(sName,'/') ;
    if (s == NULL)  {
        s = strrchr(sName,'\\') ;
        if (s == NULL)  {
            s = strrchr(sName,':') ;
        }
    }
    if (s != NULL)  {       /* had a path */
        iIndex = (s-sName)+1 ;
        strncpy(sFilePath,sName,iIndex) ;
    }
    hFind = FindFirstFile(sName,&FindData) ;
    if (hFind == INVALID_HANDLE_VALUE)  {
        fprintf(stderr,"Can't open %s\n",sName) ;
    }
    else  {
        do  {
            strcpy(&sFilePath[iIndex],FindData.cFileName) ;
            fp = fopen(sFilePath,"r") ;
            if (fp == NULL)  {
                perror(sFilePath) ;
            }
            else  {
                iGrepFile(sFilePath,fp) ;
                fclose(fp) ;
            }
        } while (FindNextFile(hFind,&FindData)) ;
        FindClose(hFind) ;
    }
}

main(
    int argc,
    char    *argv[]
){
    int     i ;
    int     c ;
    int     iFlags = REG_EXTENDED ;
    int     bFlagh = FALSE ;
    char    *sFileList = NULL ;
    char    *s ;
    char    sFileName[512] ;
    FILE    *fp ;

#define CMD_SWITCHES "cf:hinv"
    while ((c = getopt(argc,argv,CMD_SWITCHES)) != EOF)  {
        switch (c)  {
            case 'c':  {
                bCountOnly = TRUE ;
                break ;
            }
            case 'f':  {
                sFileList = optarg ;
                break ;
            }
            case 'h':  {
                bFlagh  = TRUE ;
                break ;
            }
            case 'i':  {
                iFlags |= REG_ICASE ;
                break ;
            }
            case 'n':  {
                bShowLineNumber = TRUE ;
                break ;
            }
            case 'v':  {
                bNotMatched = TRUE ;
                break ;
            }
            default:
            case '?':  {
Usage: ;
                fprintf(stderr,
                    "Usage:\n\tgrep [-" CMD_SWITCHES "] \"pattern\" [file...]\n") ;
                return(10) ;
            }
        }
    }
    if (optind >= argc)  {
        goto Usage ;
    }
    if (regcomp(&Pattern,argv[optind],iFlags) != 0)  {  /* error */
        fprintf(stderr,"Bogus pattern - \"%s\"\n",argv[optind]) ;
        return(3) ;
    }
    optind++ ;
    fp = NULL ;
    if (sFileList != NULL)  {
        if (strcmp(sFileList,"-") == 0)  {  /* from stdin */
            fp = stdin ;
        }
        else  {
            fp = fopen(sFileList,"r") ;
            if (fp == NULL)  {
                perror(sFileList) ;
                return(3) ;
            }
        }
        bMultiFile = TRUE ;         /* best guess... */
    }
    if (optind >= argc && sFileList == NULL)  {     /* no files - grep stdin */
        iGrepFile("-",stdin) ;
    }
    else  {
        if ((optind+1) < argc)  {       /* more than one file */
            bMultiFile = TRUE ;
        }
        else  {
            for (i=optind ; i<argc ; i++)  {
                if (strchr(argv[i],'*') != NULL || strchr(argv[i],'?') != NULL)  {
                    bMultiFile = TRUE ;
                    break ;
                }
            }
        }
        if (bMultiFile)  {
            if (!bFlagh)  {
                bShowFileName = TRUE ;
            }
        }
        if (sFileList != NULL)  {       /* already have it open.. */
            while (fgets(sFileName,sizeof sFileName,fp) != NULL)  {
                if ((s=strchr(sFileName,'\n')) != NULL)  {
                    *s = 0 ;
                }
                vGrepWild(sFileName) ;
            }
        }
        else  {
            for (i=optind ; i<argc ; i++)  {
                vGrepWild(argv[i]) ;
            }
        }
    }
    if (bCountOnly && bMultiFile)  {
        printf("%d lines matched (%d file(s) searched)\n",iMatches,iFiles) ;
    }
    return(0) ;
}
