/* SPAWN.C - Spawn command and yield PID/output. */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

static HANDLE
hDuplicate(
    HANDLE  hSource,
    BOOLEAN bInherit,
    DWORD   dwOptions,
    DWORD   dwAccess
){
    HANDLE  hResult ;
    HANDLE  hMyProcess = GetCurrentProcess() ;

    if (DuplicateHandle(hMyProcess,hSource,hMyProcess,&hResult,
        dwAccess,bInherit,dwOptions))  {
        return(hResult) ;
    }
    return(INVALID_HANDLE_VALUE) ;
}

static BOOLEAN
bCreateCmdProcess(
    char    *sCommand,
    HANDLE  hStdout,
    PROCESS_INFORMATION *pProcInfo
){
    STARTUPINFO StartupInfo ;
    BOOLEAN bSpawnOk ;

    ZeroMemory(&StartupInfo,sizeof StartupInfo) ;
    StartupInfo.cb = sizeof StartupInfo ;

    StartupInfo.hStdInput = hDuplicate(GetStdHandle(STD_INPUT_HANDLE),TRUE,
        DUPLICATE_SAME_ACCESS,0) ;

    StartupInfo.hStdOutput = hDuplicate(hStdout,TRUE,
        DUPLICATE_SAME_ACCESS,0) ;

    StartupInfo.hStdError = StartupInfo.hStdOutput ;

    StartupInfo.dwFlags = STARTF_USESTDHANDLES ;

    bSpawnOk = CreateProcess(
        NULL,       // application
        sCommand,   // command line
        NULL,       // process attributes
        NULL,       // thread attributes
        TRUE,       // inherit handles
        CREATE_SUSPENDED,    // creation flags
        NULL,       // environment
        NULL,       // current directory
        &StartupInfo,
        pProcInfo
        ) ;
    if (!bSpawnOk)  {
        fprintf(stderr,"CreateProcess error %lu\n",GetLastError()) ;
    }
    /*
     * now close the handles we passed to the child process.
     */
    CloseHandle(StartupInfo.hStdInput) ;
    CloseHandle(StartupInfo.hStdOutput) ;
    return(bSpawnOk) ;
}

int
main(
    int argc,
    char *argv[]
){
    PROCESS_INFORMATION ProcInfo ;
    DWORD   dwExitStatus ;
    DWORD   dwTimeout ;

    SetErrorMode(
        SEM_FAILCRITICALERRORS |
        SEM_NOGPFAULTERRORBOX  |
        SEM_NOOPENFILEERRORBOX
    ) ;
    if (argc < 2)  {
        fprintf(stderr,"Usage:\n\tspawn \"command\" [timeoutsecs]\n") ;
        return(255) ;
    }
    if (bCreateCmdProcess(argv[1],
        GetStdHandle(STD_OUTPUT_HANDLE),&ProcInfo))  {
#if 0
        printf("pid %lu\n",ProcInfo.dwProcessId) ;
#endif
        fflush(stdout) ;
        ResumeThread(ProcInfo.hThread) ;
        CloseHandle(ProcInfo.hThread) ;
        if (argc > 2)  {
            dwTimeout = atoi(argv[2])*1000L ;
        }
        else  {
            dwTimeout = INFINITE ;
        }
        WaitForSingleObject(ProcInfo.hProcess,dwTimeout) ;
        GetExitCodeProcess(ProcInfo.hProcess,&dwExitStatus) ;
        return(dwExitStatus) ;
    }
    else  {
#if 0
        printf("error %lu\n",GetLastError()) ;
#endif
        return(255) ;
    }
}
